package sets;


public class FineSet<T> extends SequentialSet<T>
{
    public FineSet() {
        super();
    }

    // find with hand-over-hand locking, return locked position
    @Override
    protected Position<T> find(Node<T> start, int key) {
        Node<T> pred, curr; // predecessor and current node in iteration
        pred = start;
        pred.lock();  // from `start' node
        curr = start.next();
        curr.lock();
        while (curr.key() < key) {
            pred.unlock();                   // unlock pred node
            pred = curr;
            curr = curr.next(); // move to next node
            curr.lock();                     // lock next node
        } // until curr.key >= key
        return new Position<T>(pred, curr);
    }

    // add while locking found position
    @Override
    public boolean add(T item) {
        // new node
        Node<T> node = newNode(item);
        Node<T> pred = null, curr = null; // dummy initialization (to make compiler happy)
        try {
            // first node from head whose key >= `node.key'
            Position<T> where = find(head, node.key());
            pred = where.pred;
            curr = where.curr;
            // pred.key < node.key <= curr.key
            return rawAdd(pred, curr, node);
        } finally {
            pred.unlock();
            curr.unlock();
        }
    }

    // remove while locking found position
    @Override
    public boolean remove(T item) {
        int key = item.hashCode();
        Node<T> pred = null, curr = null; // dummy initialization (to make compiler happy)
        try {
            // first node from head whose key >= `node.key'
            Position<T> where = find(head, key);
            pred = where.pred;
            curr = where.curr;
            // pred.key < node.key <= curr.key
            return rawRemove(pred, curr, key);
        } finally {
            pred.unlock();
            curr.unlock();
        }
    }

    // has while locking found position
    @Override
    public boolean has(T item) {
        int key = item.hashCode();
        Node<T> pred = null, curr = null; // dummy initialization (to make compiler happy)
        try {
            // first node from head whose key >= `node.key'
            Position<T> where = find(head, key);
            pred = where.pred;
            curr = where.curr;
            // pred.key < node.key <= curr.key
            return rawHas(curr, key);
        } finally {
            pred.unlock();
            curr.unlock();
        }
    }

    // node that can be individually locked
    @Override
    protected Node<T> newNode(T item) {
        return new LockableNode<>(item);
    }

    // node that can be individually locked
    @Override
    protected Node<T> newNode(int key) {
        return new LockableNode<>(key);
    }
}
